"""
This is a temp thing to override ResourceManager
"""

try:
    from pyvisa import ResourceManager
    from mcculw import ul
    from mcculw.enums import InterfaceType
except FileNotFoundError:
    raise FileNotFoundError('Please check the readme file and install the required dependencies (UL) or try running pip install mcculw')

class PiecManager():
    """
    Basically Resource Manager that melds MCC digilent stuff into it
    Alllows to get all resources
    """
    def __init__(self):
        self.rm = ResourceManager()

    def list_resources(self):
        """
        Runs list_resources() and then tries to idn each resource
        want what drivers can be used for it as well
        """
        visa = self.rm.list_resources()
        try:
            mcc = list_mcc_resources()
        except:
            print('Warning MCCULW not installed')
            mcc = []
        return tuple(list(visa) + mcc)

    def list_open_resources(self):
        visa = self.rm.list_opened_resources()
        return tuple(list(visa))
        

"""
Helper Functions
"""
def list_mcc_resources():
    ul.ignore_instacal()
    devices = ul.get_daq_device_inventory(InterfaceType.ANY)
    formatted_list = []
    if devices:
        #print('Found', len(devices), 'MCC DAQ device(s):')
        for device in devices:
            formatted_list.append(str(device.product_name) + ' (' + str(device.unique_id) + ') - '+ 'Device ADDRESS = '+ str(device.product_id))
    return formatted_list

        
        


"""
Set's up the arduino instrument class that all arduino instruments will inherit basic functionlity from.

"""

from piec.drivers.instrument import Instrument
import time
import re
import pyvisa
from pyvisa import ResourceManager
class Arduino_Stepper(Instrument):
    """
    Custom Class for using an arduino as a stepper motor driver. See https://github.com/ElPsyKurisu/STFMR/tree/main/Arduino for more information
    """
    def __init__(self, address):
        rm = ResourceManager()
        self.instrument = rm.open_resource(address, baud_rate=115200)
        self.instrument.timeout = 10000 #10s

    def idn(self):
        line = self.instrument.query("0,99") #calls in builtin method to check if serial communication works
        start_time = time.time()
        try:
            while True:
                # Check if 5 seconds have passed
                if time.time() - start_time > 5:
                    raise pyvisa.errors.VisaIOError(pyvisa.constants.VI_ERROR_TMO)
                
                # Read a line from the Arduino
                line = self.instrument.read().strip()
                
                # Check if the line contains the "Connected" message
                if "Connected" in line:
                    return "Custom Arduino_Stepper Object at {}".format(self.instrument.resource_name)
                
        except pyvisa.errors.VisaIOError as e: 
            if e.error_code == pyvisa.constants.VI_ERROR_TMO:
                print("Timeout error occurred while waiting for the Arduino.")
                return "Not connected"
    
    def step(self, num_steps, direction):
        """
        Steps the stepper motor 
        args:
            self (pyvisa.resources.asrl.not sure): Arduino
            num_steps (str/int): Desired step size, must be an integer value
            direction (str/int): [0,1] are the ONLY allowed values, 0 for CW 1 for CCW (not could be backwards)
        returns:
            current_position (int) The current position as read from the arduino
        """
        answer =self.instrument.query("{},{}".format(num_steps, direction)) #specially formatted string for arduino code to work. See https://github.com/ElPsyKurisu/STFMR/tree/main/Arduino for more information
        try:
            number = int(re.search(r'-?\d+', answer).group())
                
            # Check if the line contains the "complete" message
            if "Complete" in answer:
                return number
                
        except pyvisa.errors.VisaIOError as e: 
            if e.error_code == pyvisa.constants.VI_ERROR_TMO:
                print("Timeout error occurred while waiting for the Arduino.")


    def set_zero(self):
        """
        Hardcoded command that sets the arduinos position tracker to zero
        """
        time.sleep(2) #ensure that arduino has time to recieve the data
        self.step(0,9)

    def read_position(self):
        """
        Reads the current position without stepping the stepper
        """
        time.sleep(2) #ensure that arduino has time to recieve the data
        position = self.step(0,0) #steps 0 steps, so doesnt change position
        if position is not None:
            return position
        else:
            print("Position unknown")




"""
This is copy pasted from EKPY to allow for legacy support for certain instruments
Currently works with Measurement Computing USB-1208HS-2A0
but all similar models should work
NOTE: Does not use unified instrument scheme since doesnt support SCPI commands (i think)
"""


try:
    from mcculw import ul
    from mcculw.enums import InterfaceType
    from mcculw.device_info import DaqDeviceInfo
except FileNotFoundError:
    raise FileNotFoundError('Please check the readme file and install the required dependencies (UL) or try running pip install mcculw')
import numpy as np

__all__ = ('v_in', 'v_out', 'config_device', 'release_device')


def v_in(board_num, channel, ul_range=15, options=0):
    '''
    Wrapper for UL.v_in.
    '''
    value = ul.v_in(board_num, channel, ul_range, options)
    return value

def v_out(board_num, channel, data_value, ul_range=1, options=0):
    '''
    Wrapper for UL.v_out, dont change ul_range
    '''
    ul.v_out(board_num, channel, ul_range, data_value, options)

def release_device(board_num):
    '''
    Wrapper for ul.release_daq_device
    '''
    ul.release_daq_device(board_num)


def config_device(use_device_detection=True, dev_id_list=[],
                         board_num=0):
    '''
    Configures the device and returns the accepted params from the device
    Returns:
    list: list of channels
    int: num of channels


    '''   
    # By default, the example detects and displays all available devices and
    # selects the first device listed. Use the dev_id_list variable to filter
    # detected devices by device ID (see UL documentation for device IDs).
    # If use_device_detection is set to False, the board_num variable needs to
    # match the desired board number configured with Instacal.

    try:
        if use_device_detection:
            config_first_detected_device(board_num, dev_id_list)

        daq_dev_info = DaqDeviceInfo(board_num)
        if not daq_dev_info.supports_analog_output:
            raise Exception('Error: The DAQ device does not support '
                            'analog output')

        print('\nActive DAQ device: ', daq_dev_info.product_name, ' (',
              daq_dev_info.unique_id, ')\n', sep='')

        ao_info = daq_dev_info.get_ao_info()
        ao_range = ao_info.supported_ranges[0]
        low_chan = 0
        high_chan = min(3, ao_info.num_chans - 1)
        num_chans = high_chan - low_chan + 1

    except Exception as e:
        print('\n', e)
    return ao_range, num_chans

'''
Helper Functions taken directly from mcculw examples library
'''

def config_first_detected_device(board_num, dev_id_list=None):
    """Adds the first available device to the UL.  If a types_list is specified,
    the first available device in the types list will be add to the UL.

    Parameters
    ----------
    board_num : int
        The board number to assign to the board when configuring the device.

    dev_id_list : list[int], optional
        A list of product IDs used to filter the results. Default is None.
        See UL documentation for device IDs.
    """
    ul.ignore_instacal()
    devices = ul.get_daq_device_inventory(InterfaceType.ANY)
    if not devices:
        raise Exception('Error: No DAQ devices found')

    print('Found', len(devices), 'DAQ device(s):')
    for device in devices:
        print('  ', device.product_name, ' (', device.unique_id, ') - ',
              'Device ID = ', device.product_id, sep='')

    device = devices[0]
    if dev_id_list:
        device = next((device for device in devices
                       if device.product_id in dev_id_list), None)
        if not device:
            err_str = 'Error: No DAQ device found in device ID list: '
            err_str += ','.join(str(dev_id) for dev_id in dev_id_list)
            raise Exception(err_str)

    # Add the first DAQ device to the UL with the specified board number
    ul.create_daq_device(board_num, device)


def add_example_data(board_num, data_array, ao_range, num_chans, rate,
                     points_per_channel):
    # Calculate frequencies that will work well with the size of the array
    frequencies = []
    for channel_num in range(num_chans):
        frequencies.append(
            (channel_num + 1) / (points_per_channel / rate) * 10)

    # Calculate an amplitude and y-offset for the signal
    # to fill the analog output range
    amplitude = (ao_range.range_max - ao_range.range_min) / 2
    y_offset = (amplitude + ao_range.range_min) / 2

    # Fill the array with sine wave data at the calculated frequencies.
    # Note that since we are using the SCALEDATA option, the values
    # added to data_array are the actual voltage values that the device
    # will output
    data_index = 0
    for point_num in range(points_per_channel):
        for channel_num in range(num_chans):
            freq = frequencies[channel_num]
            value = amplitude * np.sin(2 * np.pi * freq * point_num / rate) + y_offset
            raw_value = ul.from_eng_units(board_num, ao_range, value)
            data_array[data_index] = raw_value
            data_index += 1

    return frequencies





'''
This is for the EDC Model 522 NOTE: Does not take SCPI commands
'''
import numpy as np
from piec.drivers.instrument import Instrument
#yes

class EDC522(Instrument):
    """
    Specific Class for exact model of Model 522. Currently only supports a single read out funciton
    """
    voltage_range = (-100, 100) #volts
    current_range = (-.1, .1) #amps


    def idn(self):
        """Query the instrument identity."""
        self.instrument.write("ID?")
        return self.instrument.read()

    def query_error(self):
        """Query the instrument for any error messages."""
        self.instrument.write("?")
        return self.instrument.read()
     

    def set_output(self, value, mode="voltage", opt=False):
        """
        Formats a current or voltage value into an 8-character string for instrument control.
        Automatically determines the appropriate range.

        Args:
            value (float or int): The value to send to the instrument.
            mode (str, optional): "voltage" or "current" or "crowbar". Defaults to "voltage".
            opt (bool, optional): Is only TRUE if high voltage option is connected. Enables the 1000V range

        Returns:
            str: An 8-character command string, or None if input is invalid or value is out of range.
        """
        if opt:
            self.voltage_range = (-1000, 1000)
        if mode not in ("voltage", "current", "crowbar"):
            return None
        if mode == "crowbar":
            self.instrument.write("00000000") #puts it in crowbar mode
        if value == 0:
            self.instrument.write("+0000000") #defaults to positive mode
            return
        if value == "-0":
            self.instrument.write("-0000000") #negative negative mode
            return

        polarity = "+" if value > 0 else "-"
        abs_value = abs(value)

        if mode == "voltage":
            if abs_value >max(self.voltage_range):
                raise ValueError("Voltage value out of range")
            ranges = [0.1, 10, 100, 1000]
            range_chars = "0123"
            max_values = [0.9999999, 10, 100, 1000]  # Slightly higher max values
        elif mode == "current":
            if abs_value >max(self.current_range):
                raise ValueError("Current value out of range")
            ranges = [0.01, 0.1]
            range_chars = "45"
            max_values = [0.00999999, 0.1]  # Slightly higher max values
        else:
            return None
        #NOTE, want it to defaault to lowest possible range if possible. therefore iterate bottom up
        for i, r in enumerate(ranges):
            if abs_value <= max_values[i]:  # Check against max value for the range
                scaled = abs_value / r 
                best_range_index = i
                scaled_value = scaled
                break
        else:  # No suitable range was found
            return None

        digits_str = "{:06.0f}".format(scaled_value * 1000000)

        # J handling:
        J_vals = [100e-3, 10, 100, 1000]
        if abs_value in J_vals:
            digits_str = "J00000"
        command = f"{polarity}{digits_str}{range_chars[best_range_index]}"
        print(command)
        self.instrument.write(command)

#helper func
import re

def extract_number(input_string):
    # Use regular expression to find the number in the string
    match = re.search(r'[+-]?\d+\.\d+E[+-]?\d+', input_string)
    if match:
        return match.group(0)
    else:
        return None
    
"""
Old function to allow for the range 10 and 11.1111 values

def format_instrument_command(value, data_type="voltage"):
    '''Formats a value into an 8-character instrument command string.'''

    if data_type not in ("voltage", "current"):
        return None

    if value == 0:
        return "00000000"

    polarity = "+" if value > 0 else "-"
    abs_value = abs(value)

    if data_type == "voltage":
        ranges = [0.1, 10, 100, 1000]
        range_chars = "0123"
        max_values = [0.9999999, 11.11111, 111.1111, 1111.111]  # Slightly higher max values
    elif data_type == "current":
        ranges = [0.01, 0.1]
        range_chars = "45"
        max_values = [0.00999999, 0.1111111]  # Slightly higher max values
    else:
        return None

    for i, r in enumerate(ranges):
        if abs_value <= max_values[i]:  # Check against max value for the range
            scaled = abs_value / r
            best_range_index = i
            scaled_value = scaled
            break
    else:  # No suitable range was found
        return None

    digits_str = "{:06.0f}".format(scaled_value * 100000)

    # J handling:
    if best_range_index == 1 and scaled_value == 10:  # Exactly 10V
        digits_str = "J00000"
    elif best_range_index == 1 and 1 <= scaled_value < 10: # 1V to 9.99999V in 10V range
        digits = []
        for digit in str(int(scaled_value * 100000)):
            if digit == '1':
                digits.append('J')
            else:
                digits.append(digit)
        digits_str = "".join(digits).zfill(6)

    return f"{polarity}{digits_str}{range_chars[best_range_index]}"

"""




"""
Set's up the scpi instrument class that all scpi instruments will inherit basic functionlity from to be followed by sub classes (e.g. scope, wavegen)
"""
from typing import Union
import numpy as np
import pandas as pd
from pyvisa import ResourceManager
import time
import re
import json
import os
from piec.drivers.instrument import Instrument


class VirtualRMInstrument:
    """
    This class replaces the resource manager object in the virtual case,
    just needs to replace the .write() and .query() methods
    """
    def __init__(self, verbose:bool = False):
        self.verbose = verbose
        print('INITIALIZING VIRTUAL RESOURCE MANAGER, VISA NOT CONNECTED')
        current_dir = os.path.dirname(__file__)

        # Construct the full path to the virtual query JSON file
        json_path = os.path.join(current_dir, "virtual_scpi_queries.json")

        # Load the JSON file
        with open(json_path, "r") as file:
            self.query_dict = json.load(file)

    def query(self, input:str):
        time.sleep(0.01)
        if self.verbose:
            print('Query recieved: ',input)
        try:
            return self.query_dict[input]
        except:
            print('QUERY: ', input, ' NOT IN virtual_scpi_queries.json')
            return "VIRTUAL QUERY:"+input


    def write(self, input:str):
        time.sleep(0.01)
        if self.verbose:
            print('Write recieved: ',input)
    
    def write_binary_values(self, data, scaled_data, datatype='h'):
        time.sleep(0.01)
        if self.verbose:
            print('Binary write recieved: ', data, scaled_data, datatype)

# Define a class
class SCPI_Instrument(Instrument):
    # Initializer / Instance attributes
    def __init__(self, address):
        rm = ResourceManager()
        if address == 'VIRTUAL':
            self.instrument = VirtualRMInstrument() # initiate instrument in virtual mode
        else:
            self.instrument = rm.open_resource(address) #comment out to debug without VISA connection
        self.virtual = (address=='VIRTUAL')

    def _debug(self, **args):
        """
        Debugging Function for _check_params not intended for actual code use, REQUIRES self.instrument above to be commented out
        """
        self._check_params(locals()['args']) #only need this here since i did **args

    # Generic Methods all instruments should have
    def idn(self):
        """
        Queries the instrument for its ID

        """
        return self.instrument.query("*IDN?")


    def reset(self):
        """
        Resets the instrument to its default parameters
        """
        self.instrument.write("*RST")

    def initialize(self):
        """
        Resets the instrument and clears the registry
        """
        self.instrument.write("*RST")
        self.instrument.write("*CLS")

    def operation_complete_query(self):
        """
        Asks the instrument if the operation is complete by sending the *OPC query
        Returns a 1 if complete
        Note that using the *OPC? query holds the I/O bus until the instrument commands have been completed. 
        It may not be acceptable to hold the GPIB bus when you are controlling multiple instruments. In those cases
        use the *OPC bit in the Standard Event register and implement some form of status reporting system. An example
        using the *OPC bit is provided in the Using SRQ Events section.
        """
        response = self.instrument.query("*OPC?")
        return response.strip()

    def check_errors(self):
        """
        Returns a list of errors or None if there are no errors
        Technically not an IEEE Mandated command but all instruments for our purposes should use this.
        With help from ChatGPT
        """
        errors = []
        start_time = time.time()
        while True:
            current_time = time.time()
            if current_time - start_time > 5:
                break
            error = self.instrument.query('SYST:ERR?')
            if check_error_string(error):
                break
            errors.append(error)
        if len(errors) == 0:
            errors = None
        return errors


    def print_specs(self):
        """
        Function that lists all class attributes for the instrument. NOTE: May not work in virtual case
        """
        spec_dict = get_class_attributes_from_instance(self)
        for key in spec_dict:
            if spec_dict[key] != type(dict):
                print(key, ':', spec_dict[key])
            else:
                key_dict = getattr(self, key) #this is a dict that the key is the type e.g. list or dict
                key_dict_key = list(key_dict.keys())[0]
                key_dict_key_value = key_dict[key_dict_key]
                print(key, ':', key_dict_key_value)
            
    def _check_params(self, locals_dict):
        """
        Want to check class attributes and arguments from the function are in acceptable ranges. Uses .locals() to get all arguments and checks
        against all class attributes and ensures if they match the range is valid 
        """
        class_attributes = get_class_attributes_from_instance(self)
        keys_to_check = get_matching_keys(locals_dict, class_attributes)
        for key in keys_to_check:
            attribute_value = getattr(self, key) #allowed types are strings, tuples, lists, and dicts
            if attribute_value is None:
                print("Warning no range-checking defined for \033[1m{}\033[0m, skipping _check_params".format(key)) #makes bold text
                continue
            input_value = locals_dict[key]
            if type(attribute_value) == tuple:
                if not is_value_between(input_value, attribute_value): #will error need to make jey values correct
                    exit_with_error("Error input value of \033[1m{}\033[0m for arg \033[1m{}\033[0m is out of acceptable Range \033[1m{}\033[0m".format(input_value, key, attribute_value))
            if type(attribute_value) == list:
                if not is_contained(input_value, attribute_value): #checks if the input value is in the allowed list
                    exit_with_error("Error input value of \033[1m{}\033[0m for arg \033[1m{}\033[0m is not in list of acceptable \033[1m{}\033[0m".format(input_value, key, attribute_value))
            if type(attribute_value) == dict:
                #need helper function here
                attribute_key = get_matching_keys(locals_dict, attribute_value) #this is the first key in the class attribute dict
                if len(attribute_key) != 1:
                    print("WARNING found {} keys instead of 1, skipping {} checking".format(len(attribute_key), key))
                else: #assumes input is range btw, since no other reason for using a dict i believe. WIll update if changes
                    local_value = locals_dict[attribute_key[0]] #this is the value
                    attribute_sub_dict = attribute_value[attribute_key[0]]
                    if not is_value_between(input_value, attribute_sub_dict[local_value]):
                        exit_with_error("Error input value of \033[1m{}\033[0m for arg \033[1m{}\033[0m is out of acceptable Range \033[1m{}\033[0m".format(input_value, attribute_key, attribute_sub_dict[local_value]))

class Scope(SCPI_Instrument):
    """
    Sub-class of Instrument to hold the general methods used by scopes. For Now defaulted to DSOX3024a, but can always ovveride certain SCOPE functions
    """
    #Should be overriden
    voltage_range = None #entire screen range
    voltage_scale = None #units per division
    time_range = None   
    time_scale = None
    time_base_type = None

    def setup(self, channel: str=1, voltage_range: str=16, voltage_offset: str=1.00, delay: str='100e-6',
          time_range: str='1e-3', autoscale=True):
        """
        Sets up the oscilliscope with the given paramaters. If autoscale is turned on it will ignore
        all other arguments and simply autoscale the instrument. Otherwise sample paramters are given
        as the default values. First the Program resets the instrument and after passing in desired parameters
        it sets the scope up for acquiring.

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): SCOPE
            channel (str): Desired channel allowed values are 1,2,3,4
            voltage_range (str): The y scale of the oscilloscope, max is 40V, min is 8mV
            voltage_offset (str): The offset for the voltage in units of volts
            delay (str): The delay in units of s
            time_range (str): The x scale of the oscilloscope, min 20ns, max 500s
        """
        self._check_params(locals())
        self.reset()
        if autoscale:
            self.instrument.write(":AUToscale")
        else:
            self.instrument.write("CHANel{}:RANGe {}".format(channel, voltage_range))
            self.instrument.write("CHANel{}:OFFSet {}".format(channel, voltage_offset))
            self.instrument.write("CHANel{}:TIMebase:RANGe {}".format(channel, time_range))
            self.instrument.write("CHANel{}:TIMebase:DELay {}".format(channel, delay))
        self.instrument.write(":ACQuire:TYPE NORMal")

    def configure_timebase(self, time_base_type="MAIN", position="0.0",
                       reference="CENT", time_range=None, time_scale=None, vernier=False):
        """Configures the timebase of the oscilliscope. Adapted from EKPY program 'Configure Timebase (Basic)'
        Should call initialize first.

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            time_base_type (str): Allowed values are 'MAIN', 'WINDow', 'XY', and 'ROLL', note must use main for data acquisition
            position (str): The position in the scope, [0.0] is a good default This is actually the delay on the scope (moves in time right and left)
            time_range (str): The x range of the scope min is 20ns, max is 500s
            time_scale (str): The x scale of the scope in units of s/div min is 2ns, max is 50s
            vernier (boolean): Enables Vernier scale
        """
        self._check_params(locals())
        if time_base_type is not None:
            self.instrument.write("TIM:MODE {}".format(time_base_type))
        if position is not None:
            self.instrument.write("TIM:POS {}".format(position))
        if time_range is not None:
            self.instrument.write("TIM:RANG {}".format(time_range))
        if reference is not None:
            self.instrument.write("TIM:REF {}".format(reference))
        if time_scale is not None:
            self.instrument.write("TIM:SCAL {}".format(time_scale))
        if vernier:
            self.instrument.write("TIM:VERN ON")
        else:
            self.instrument.write("TIM:VERN OFF")

    def configure_channel(self, channel: str='1', scale_mode=True, voltage_scale: str='4', voltage_range: str='40',
                              voltage_offset: str='0.0', coupling: str='DC', probe_attenuation: str='1.0', 
                              impedance: str='ONEM', enable_channel=True):
        """Sets up the voltage measurement on the desired channel with the desired paramaters. Taken from
        EKPY. 

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            channel (str): Desired channel allowed values are 1,2,3,4
            scale_mode (boolean): Allows us to select between a vertical scale or range setting [see options below]
            voltage_scale (str): The vertical scale in units of v/div
            voltage_range (str): The verticale scale range min: 8mv, max: 40V
            voltage_offset (str): The offset for the vertical scale in units of volts
            coupling (str): 'AC' or 'DC' values allowed
            probe_attenuation (str): Multiplicative factor to attenuate signal to stil be able to read, max is most likely 10:1
            impedance (str): Configures if we are in high impedance mode or impedance match. Allowed factors are 'ONEM' for 1 M Ohm and 'FIFT' for 50 Ohm
            enable_channel (boolean): Enables the channel
        """
        self._check_params(locals())
        if scale_mode:
            self.instrument.write("CHAN{}:SCAL {}".format(channel, voltage_scale))
        else:
            self.instrument.write("CHAN{}:RANG {}".format(channel, voltage_range))
        self.instrument.write("CHAN{}:OFFS {}".format(channel, voltage_offset))
        self.instrument.write("CHAN{}:COUP {}".format(channel, coupling))
        self.instrument.write("CHAN{}:PROB {}".format(channel, probe_attenuation))
        self.instrument.write("CHAN{}:IMP {}".format(channel, impedance))
        if enable_channel:
            self.instrument.write("CHAN{}:DISP ON".format(channel))
        else:
            self.instrument.write("CHAN{}:DISP OFF".format(channel))
    
    def configure_trigger_characteristics(self, type: str='EDGE', holdoff_time: str='4E-8', low_voltage_level: str='1',
                                      high_voltage_level: str='1', trigger_source: str='CHAN1', sweep: str='AUTO',
                                       enable_high_freq_filter=False, enable_noise_filter=False):
        """Configures the trigger characteristics Taken from EKPY. 'Configures the basic settings of the trigger.'
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            type (str): Trigger type, accepted params are: [EDGE (Edge), GLIT (Glitch), PATT (Pattern), TV (TV), EBUR (Edge Burst), RUNT (Runt), NFC (Setup Hold), TRAN (Transition), SBUS1 (Serial Bus 1), SBUS2 (Serial Bus 2), USB (USB), DEL (Delay), OR (OR), NFC (Near Field Communication)]
            holdoff_time (str): Additional Delay in units of sec before re-arming trigger circuit
            low_voltage_level (str): The low trigger voltage level units of volts
            high_voltage_level (str): The high trigger voltage level units of volts
            trigger_source (str): Desired channel to trigger on allowed values are [CHAN1,CHAN2,CHAN3,CHAN4, EXT (there are more)]
            sweep (str): Allowed values are [AUTO (automatic), NORM (Normal)]
            enable_high_freq_filter (boolean): Toggles the high frequency filter
            enable_noise_filter (boolean): Toggles the noise filter
        """
        self._check_params(locals())
        if enable_high_freq_filter:
            self.instrument.write(":TRIG:HFR ON")
        else:
            self.instrument.write(":TRIG:HFR OFF")
        self.instrument.write(":TRIG:HOLD {}".format(holdoff_time))
        self.instrument.write(":TRIG:LEV:HIGH {}, {}".format(high_voltage_level, trigger_source))
        self.instrument.write(":TRIG:LEV:LOW {}, {}".format(low_voltage_level, trigger_source))
        self.instrument.write(":TRIG:MODE {}".format(type))
        if enable_noise_filter:
            self.instrument.write(":TRIG:NREJ ON")
        else:
            self.instrument.write(":TRIG:NREJ OFF")
        self.instrument.write(":TRIG:SWE {}".format(sweep))

    def configure_trigger_edge(self, trigger_source: str='CHAN1', input_coupling: str='AC', edge_slope: str='POS', 
                           level: str='0', filter_type: str='OFF'):
        """Configures the trigger characteristics Taken from EKPY. 'Configures the basic settings of the trigger.'
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            trigger_source (str): Desired channel/source to trigger on allowed values are: [CHAN1,CHAN2,CHAN3,CHAN4,DIG0,DIG1 (there are more)]
            input_coupling (str): Allowed values = [AC, DC, LFR (Low Frequency Coupling)]
            edge_slope (str): Allowed values = [POS, NEG, EITH (either), ALT (alternate)]
            level (str): Trigger level in volts
            filter_type (str): Allowed values = [OFF, LFR (High-pass filter), HFR (Low-pass filter)] Note: Low Frequency reject == High-pass
        """
        self._check_params(locals())
        self.instrument.write(":TRIG:SOUR {}".format(trigger_source))
        self.instrument.write(":TRIG:COUP {}".format(input_coupling))
        self.instrument.write(":TRIG:LEV {}".format(level))
        self.instrument.write(":TRIG:REJ {}".format(filter_type))
        self.instrument.write(":TRIG:SLOP {}".format(edge_slope))

    def initiate(self):
        """
        The :DIGitize command is a specialized RUN command. It causes the instrument 
    to acquire waveforms according to the settings of the :ACQuire commands 
    subsystem. When the acquisition is complete, the instrument is stopped.
    If no argument is given, :DIGitize acquires the channels currently displayed. If no 
    channels are displayed, all channels are acquired.

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
        """
        self.instrument.write(":DIG")
        self.instrument.write("*CLS")
    
    def setup_wf(self, source: str='CHAN1', byte_order: str='MSBF', format: str='byte', points: str='1000', 
             points_mode: str='NORMal', unsigned: str='OFF'):
        """Sets up the waveform with averaging or not and of a specified format/count  
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            source (str): Desired channel allowed values are [CHAN1, CHAN2, CHAN3, CHAN4, FUNC, SBUS1, etc]
            byte_order (str): Either MSBF (most significant byte first) or LSBF (least significant byte first)
            format (str): Format of data allowed args are [ASCii (floating point), WORD (16bit two-bytes), BYTE (8-bit)]
            points (str): Number of data points for the waveform to return allowed args are [100,250,500,1000] for NORM or up to [8000000] for MAX or RAW
            points_mode (str): Mode for points allowed args are [NORM (normal), MAX (maximum), RAW]
            unsigned (str): Allows to switch between unsigned and signed integers [OFF (signed), ON (unsigned)]
        """
        self._check_params(locals())
        self.instrument.write(":WAVeform:SOURce {}".format(source))
        self.instrument.write(":WAVeform:BYTeorder {}".format(byte_order))
        self.instrument.write(":WAVeform:FORMat {}".format(format))
        self.instrument.write(":WAVeform:POINts:MODE {}".format(points_mode))
        self.instrument.write(":WAVeform:POINts {}".format(points))
        self.instrument.write(":WAVeform:UNSigned {}".format(unsigned))

    def query_wf(self, byte_order: str='MSBF', unsigned: str='OFF'):
        """Returns the specified channels waveform with averaging or not and of a specified format/count, call
        setup_wf first to intialize it correctly. This function only calls queries. First calls preamble to get
        data format. Then parses data and converts data to usable format.
        GET_PREAMBLE - The preamble block contains all of the current
        ' WAVEFORM settings. It is returned in the form <preamble_block><NL>
        ' where <preamble_block> is:
        ' FORMAT : int16-0= BYTE, 1 = WORD, 4 = ASCII.
        ' TYPE : int16-0= NORMAL, 1 = PEAK DETECT, 2 = AVERAGE
        ' POINTS : int32 - number of data points transferred.
        ' COUNT : int32 - 1 and is always 1.
        ' XINCREMENT : float64 - time difference between data points.
        ' XORIGIN : float64 - always the first data point in memory.
        ' XREFERENCE : int32 - specifies the data point associated with
        ' x-origin.
        ' YINCREMENT : float32 - voltage diff between data points.
        ' YORIGIN : float32 - value is the voltage at center screen.
        ' YREFERENCE : int32 - specifies the data point where y-origin
        ' occurs 

        First it calls operation complete to ensure its finished taking data
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight DSOX3024a
            byte_order (str): Either MSBF (most significant byte first) or LSBF (least significant byte first)
            unsigned (str): Allows to switch between unsigned and signed integers [OFF (signed), ON (unsigned)]

        returns:
            preamble_dict (dict) Dictionary with all params labeled. (MetaData)
            time (list): Python list with all the scaled time (x_data array)
            wfm (list): Python list with all the scaled y_values (y_data array) 
        """
        self._check_params(locals())
        check = self.operation_complete_query() #add error handling if not returns 1
        if check != '1':
            exit_with_error("Scope Not Ready: Operation Complete Query Failed")
        preamble = self.instrument.query(":WAVeform:PREamble?")
        preamble1 = preamble.split()
        preamble_list = preamble1[0].split(',')
        preamble_dict = {
        'format': np.int16(preamble_list[0]),
        'type': np.int16(preamble_list[1]),
        'points': np.int32(preamble_list[2]),
        'count': np.int32(preamble_list[3]),
        'x_increment': np.float64(preamble_list[4]),
        'x_origin': np.float64(preamble_list[5]),
        'x_reference': np.int32(preamble_list[6]),
        'y_increment': np.float32(preamble_list[7]),
        'y_origin': np.float32(preamble_list[8]),
        'y_reference': np.int32(preamble_list[9]),
        }
        if byte_order == 'MSBF':
            is_big_endian = True
        if byte_order == 'LSBF':
            is_big_endian = False
        if unsigned == 'OFF':
            is_unsigned = False
        if unsigned == 'ON':
            is_unsigned = True
        if self.virtual:
            data_df = pd.read_csv(os.path.join(os.path.dirname(__file__), "virtual_osc_trace.csv"))
            time = data_df['time (s)'].values
            wfm = data_df['voltage (V)'].values
        else:
            if preamble_dict["format"] == 0 and not is_unsigned:
                data = self.instrument.query_binary_values("WAVeform:DATA?", datatype='b', is_big_endian=is_big_endian)
            if preamble_dict["format"] == 0 and is_unsigned:
                data = self.instrument.query_binary_values("WAVeform:DATA?", datatype='B', is_big_endian=is_big_endian)
            if preamble_dict["format"] == 1 and not is_unsigned:
                data = self.instrument.query_binary_values("WAVeform:DATA?", datatype='h', is_big_endian=is_big_endian)
            if preamble_dict["format"] == 1 and is_unsigned:
                data = self.instrument.query_binary_values("WAVeform:DATA?", datatype='H', is_big_endian=is_big_endian)
            if preamble_dict["format"] == 4:
                data = self.instrument.query_ascii_values("WAVeform:DATA?")
            time = []
            wfm = []
            for t in range(preamble_dict["points"]):
                time.append((t* preamble_dict["x_increment"]) + preamble_dict["x_origin"])
            for d in data:
                wfm.append((d * preamble_dict["y_increment"]) + preamble_dict["y_origin"])
        return preamble_dict, time, wfm


class Awg(SCPI_Instrument):
    """
    Sub-class of Instrument to hold the general methods used by an awg. For Now defaulted to keysight81150a, but can always ovveride certain SCOPE functions
    """
    #Should be overriden
    channel = None
    voltage = None
    frequency = None
    func = None #might be useless since all awgs should have sin, squ, pulse etc
    slew_rate = None #1V/ns

    def configure_impedance(self, channel: str='1', source_impedance: str='50.0', load_impedance: str='50.0'):
        """
        This program configures the output and input impedance of the wavegen. Taken from EKPY.
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            source_impedance (str): The desired source impedance in units of Ohms, allowed args are [5, 50]
            load_impedance (str): The desired load impedance in units of Ohms, allowed args are [0.3 to 1E6]

        """
        self.instrument.write(":OUTP{}:IMP {}".format(channel, source_impedance))
        #wavegen.write(":OUTP{}:LOAD {}".format(channel, load_impedance)) Also valid for below
        self.instrument.write(":OUTP{}:IMP:EXT {}".format(channel, load_impedance))

    def configure_trigger(self, channel: str='1', trigger_source: str='IMM', mode: str='EDGE', slope: str='POS'):
        """
        This program configures the trigger. Taken from EKPY.
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            trigger_source (str): Trigger source allowed args = [IMM (immediate), INT2 (internal), EXT (external), MAN (software trigger)]
            mode (str): The type of triggering allowed args = [EDGE (edge), LEV (level)]
            slope (str): The slope of triggering allowed args = [POS (positive), NEG (negative), EIT (either)]
        """ 
        self.instrument.write(":ARM:SOUR{} {}".format(channel, trigger_source))
        self.instrument.write(":ARM:SENS{} {}".format(channel, mode))
        self.instrument.write(":ARM:SLOP {}".format(slope))

    def create_arb_wf(self, data: Union[np.array, list], name=None, channel='1'):
        """
        NOTE: DOES NOT SCALE HORIZONTALLY YET-> aka wont fill out to 524288 points
        This program creates an arbitrary waveform within the limitations of the
        Keysight 81150A which has a limit of 2 - 524288 data points. In order to send data
        in accordance with the 488.2 block format which looks like #ABC, where '#' marks the start
        of the data flow and 'A' refers to the number of digits in the byte count, 'B' refers to the
        byte count and 'C' refers to the actual data in binary. The data is first scaled between
        -8191 to 8191 in accordance to our instrument. Adapted from EKPY (note their arrays
        contain 10,000 elements). 
        Note: Will NOT save waveform in non-volatile memory if all the user available slots are
        filled (There are 4 allowed at 1 time plus 1 in volatile memory).

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            data (ndarray or list): Data to be converted to wf
            name (str): Name of waveform, must start with A-Z
            channel (str): What channel to put the volatile WF on
        """  
        # Ensure data is a numpy array
        data = np.array(data)
        #check length of data is valid
        dict_to_check = locals()
        dict_to_check['arb_wf_points_range'] = len(data) #this adds to our _check_params the class attribute 'arb_wf_points_range'
        self._check_params(dict_to_check)

        # Scale the waveform data to the valid range See scale_waveform_data
        scaled_data = scale_waveform_data(data)  
        self.instrument.write(":FORM:BORD SWAP")

        self.instrument.write_binary_values(":DATA{}:DAC VOLATILE, ".format(channel), scaled_data, datatype='h') #need h as 2bit bytes see struct module
        if name is not None:
            #first check if has room to copy
            slots_available = self.instrument.query('DATA:NVOLatile:FREE?').strip() #returns a number corresponding to num_slots_free
            if int(slots_available) == 0:
                stored_wfs = self.instrument.query('DATA:NVOLatile:CATalog?').strip() #checks the stored_wfs in voltatile memory
                stored_wfs_list = stored_wfs.replace('"', '').split(',')
                name_to_delete = ask_user_to_select(stored_wfs_list)
                self.instrument.write(":DATA:DEL {}".format(name_to_delete))

            self.instrument.write(":DATA:COPY {}, VOLATILE".format(name))

    def create_arb_wf_legacy(self, data, name=None):
        """
        NOTE THIS IS SUPERCEDED BY THE BINARY INTERPRETATION
        This program creates an arbitrary waveform using the slow non binary format, see create_arbitrary_wf_binary for more info
        NOTE: Will NOT save waveform in non-volatile memory, unless a name is given.
        NOTE: Will NOT save waveform in non-volatile memory if all the user available slots are
        filled (There are 4 allowed at 1 time plus 1 in volatile memory).
        Also for 10k points it is quite slow, allow for like 3 seconds to send the data. Will need to rewrite the binary version
        if we want speed

        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            data (ndarray or list): Data to be converted to wf
            name (str): Name of waveform, must start with A-Z
        """
        data_string = ""
        for i in range(len(data)):
            data_string += str(data[i]) +','
        data_string = data_string[:-1] #remove last comma
        self.instrument.write(":DATA VOLATILE, {}".format(data_string))
        if name is not None:
            self.instrument.write(":DATA:COPY {}, VOLATILE".format(name))

    def configure_wf(self, channel: str='1', func: str='SIN', voltage: str='1.0', offset: str='0.00', frequency: str='1e3', duty_cycle='50',
                      num_cycles=None, invert: bool=False):
        """
        This function configures the named func with the given parameters. Works on both user defined and built-in functions
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            func (str): The function name as saved on the instrument
            voltage (str): The V_pp of the waveform in volts
            offset (str): The voltage offset in units of volts
            frequency (str): the frequency in units of Hz for the arbitrary waveform
            duty_cycle (str): duty_cycle defined as 100* pulse_width / Period ranges from 0-100, (cant actually do 0 or 100 but in between is fine)
            num_cycles (str): number of cycles by default set to None which means continous NOTE only works under BURST mode, not implememnted
            invert (bool): Inverts the waveform by flipping the polarity
        """
        #might need to rewrite check_params here
        #self._check_params(locals()) #this wont work for user defined functions...
        built_in_list = ['SIN', 'SQU', 'RAMP', 'PULS', 'NOIS', 'DC'] #check if built in, else use checkparams or this should be via check_params
        user_funcs = self.instrument.query(":DATA:CAT?")
        user_funcs_list = user_funcs.replace('"', '').split(',')
        if func in built_in_list:
            self._configure_built_in_wf(channel, func, frequency, voltage, offset, duty_cycle)
        else:
            self._configure_arb_wf(channel, func, voltage, offset, frequency, invert)

    def _configure_built_in_wf(self, channel: str='1', func='SIN', frequency='1e3', voltage='1', offset='0', duty_cycle='50', invert: bool=False):
        """
        Decides what built-in wf to send - by default sin

        args:
            self (pyvisa.resources.ENET-Serial INSTR): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            func (str): Desired output function, allowed args are [SIN (sinusoid), SQU (square), RAMP, PULSe, NOISe, DC, USER (arb)]
            frequency (str): frequency in Hz (have not added suffix funcitonaility yet)
            voltage (str): The V_pp of the waveform in volts
            offset (str): DC offset for waveform in volts
            duty_cycle (str): duty_cycle defined as 100* pulse_width / Period ranges from 0-100, (cant actually do 0 or 100 but in between is fine)
            num_cycles (str): number of cycles by default set to None which means continous NOTE only works under BURST mode, not implememnted
            invert (bool): Inverts the waveform by flipping the polarity
        """
        self._check_params(locals())
        self.instrument.write(":SOUR:FUNC{} {}".format(channel, func)) 
        self.instrument.write(":SOUR:FREQ{} {}".format(channel, frequency))
        self.instrument.write(":VOLT{}:OFFS {}".format(channel, offset))
        self.instrument.write(":VOLT{} {}".format(channel, voltage))
        if func.lower() == 'squ' or func.lower() == 'square':
            self.instrument.write(":SOUR:FUNC{}:SQU:DCYC {}".format(channel, duty_cycle)) 
        if func.lower() == 'pulse' or func.lower() == 'puls':
            self.instrument.write(":SOUR:FUNC{}:PULS:DCYC {}".format(channel, duty_cycle))
        if invert:
            self.instrument.write(":OUTP{}:POL INV".format(channel))
        else:
            self.instrument.write(":OUTP{}:POL NORM".format(channel))

    def _configure_arb_wf(self, channel: str='1', name='VOLATILE', voltage: str='1.0', offset: str='0.00', frequency: str='1000', invert: bool=False):
        """
        This program configures arbitrary waveform already saved on the instrument. Adapted from EKPY. 
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            name (str): The Arbitrary Waveform name as saved on the instrument, by default VOLATILE
            voltage (str): The V_pp of the waveform in volts
            offset (str): The voltage offset in units of volts
            frequency (str): the frequency in units of Hz for the arbitrary waveform
            invert (bool): Inverts the waveform by flipping the polarity
        """
        dict_to_check = locals()
        dict_to_check['func'] = 'USER' #this is useless i want to make sure frequency is good tho for arb waveform
        self._check_params(dict_to_check)
        if self.slew_rate is not None:
            points = self.instrument.query(":DATA:ATTR:POIN? {}".format(name)).strip()
            if (float(voltage))/(float(frequency)/float(points)) > self.slew_rate:
                    print('WARNING: DEFINED WAVEFORM IS FASTER THAN AWG SLEW RATE')
        self.instrument.write(":FUNC{}:USER {}".format(channel, name)) #makes current USER selected name, but does not switch instrument to it
        self.instrument.write(":FUNC{} USER".format(channel)) #switches instrument to user waveform
        self.instrument.write(":VOLT{} {}".format(channel, voltage))
        self.instrument.write(":FREQ{} {}".format(channel, frequency))
        self.instrument.write(":VOLT{}:OFFS {}".format(channel, offset))
        if invert:
            self.instrument.write(":OUTP{}:POL INV".format(channel))
        else:
            self.instrument.write(":OUTP{}:POL NORM".format(channel))


    def output_enable(self, channel: str='1', on=True):
        """
        This program toggles the selected output. Taken from EKPY. 
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            channel (str): Desired Channel to configure accepted params are [1,2]
            on (boolean): True for on, False for off
        """
        if on:
            self.instrument.write(":OUTP{} ON".format(channel))
        else:
            self.instrument.write(":OUTP{} OFF".format(channel))

    def display_enable(self, on=True):
        """
        This program toggles the display On or OFF, it is recommended for programming speed to disale the display
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
            on (boolean): True for display on, False for off
        """
        if on:
            self.instrument.write("DISP ON")
        else:
            self.instrument.write("DISP OFF")

    def send_software_trigger(self):
        """
        This program sends the software trigger. Taken from EKPY. 
        args:
            self (pyvisa.resources.gpib.GPIBInstrument): Keysight 81150A
        """
        self.instrument.write(":TRIG")

    def stop(self):
        """Stop the awg.

        args:
            self (pyvisa.resources.ENET-Serial INSTR): Keysight 81150A
        """
        self.output_enable('1', False) #should change to take into account channels available from class attributes
        self.output_enable('2', False)
        
    def couple_channels(self):
        """
        Couples the channel params so Channel 1 and 2 are identical, not sure how well the outputs will sync. 
        Convention is to make changes to channel 1 now that will affect channel 2

        args:
            self (pyvisa.resources.ENET-Serial INSTR): Keysight 81150A
            
        """
        self.instrument.write(":TRACK:CHAN1 ON")

class DMM(SCPI_Instrument):
    """
    Sub-class of Instrument to hold the general methods used by a DMM.
    """
    def configure(self):
        """
        Configures the DMM to measure the correct thing
        """
        self.instrument.write(":SENS:FUNC VOLT:DC")



"""
Helper Functions Below
"""

def ask_user_to_select(options):
    """
    Helper function to format options to choose taken from help with CHATGPT
    EXAMPLE USAGE:
    # List of options
    options = ['ARB1', 'PV', 'PUND', 'DWM']

    # Ask the user to select an option
    selected_option = ask_user_to_select(options)
    print(f"You selected: {selected_option}")
    1. ARB1
    2. PV
    3. PUND
    4. DWM
    """
    # Display the options
    for i, option in enumerate(options, start=1):
        print(f"{i}. {option}")

    # Ask the user to select an option
    while True:
        try:
            choice = int(input("Enter the number of your choice: "))
            if 1 <= choice <= len(options):
                return options[choice - 1]
            else:
                print(f"Please enter a number between 1 and {len(options)}.")
        except ValueError:
            print("Invalid input. Please enter a number.")



def is_contained(value, lst):
    """
    Helper Function that checks if a string is contained within a list and ignores case sensitivity
    """
    my_string = value.lower()
    my_list = [item.lower() for item in lst]
    return my_string in my_list


def is_value_between(value, num_tuple):
    """
    Helper function that checks if the value is between allowed ranges, taken with help from ChatGPT
    """
    if value is None:
        return True
    if type(value) is str:
        value = float(value)
    if len(num_tuple) != 2:
        raise ValueError("Tuple must contain exactly two numbers")
    return num_tuple[0] <= value <= num_tuple[1]


def get_matching_keys(dict1, dict2):
    """ 
    Find the intersection of keys from both dictionaries, taken from ChatGPT
    """
    matching_keys = set(dict1.keys()).intersection(dict2.keys())
    return list(matching_keys)

def get_class_attributes_from_instance(instance):
    """
    Helper Function to get the class attributes from an instance (calls self) with help from ChatGPT
    """
    cls = instance.__class__
    attributes = {}
    for base in cls.__mro__:
        attributes.update({attr: getattr(base, attr) for attr in base.__dict__ if not callable(getattr(base, attr)) and not attr.startswith("__")})
    return attributes


def find_first_number(input_string):
    # Define a regular expression pattern to match one or more digits
    pattern = r'\d+'
    
    # Search for the pattern in the input string
    match = re.search(pattern, input_string)
    
    # If a match is found, return the matched number
    if match:
        return match.group()
    else:
        return None

def check_error_string(error_string):
    """
    Helper Function that checks if the error str matches either a 0 or No error
    returns true if no error
    """
    # Convert the string to lowercase for case-insensitive comparison
    lower_string = error_string.lower()
    error_code = find_first_number(lower_string)
    # Check if the string contains '0' or 'no error' after a comma
    if error_code == "0" or "no error" in lower_string:
        return True
    else:
        return False

def is_integer(n):
    """
    Helper function to check if a number is an intger including stuff like 5.0
    Taken with help from ChatGPT
    """
    if isinstance(n, int):
        return True
    elif isinstance(n, float):
        return n.is_integer()
    else:
        return False


'''
Helper functions for awg class:
'''
def scale_waveform_data(data: np.array, preserve_vertical_resolution: bool=False) -> np.array:
    """
    Helper function that scales values to a max of 8191 in such a way that the abs(max) is 8191
    and the rest is uniformly scaled. All VALUES SHOULD BE INTEGERS
    NOTE YOU LOSE RESOLUTION WITH THIS METHOD if preserve_vertical_resoltuion is false, but it preserves the wf shape!
    shuld print estimated lost in  PP VOLTAGE from resolution
    """
    max_abs = np.max(abs(data))
    max_inst = 8191
    scale_factor = None
    if preserve_vertical_resolution:
        scale_factor = max_inst/max_abs
    else:
        while is_integer(scale_factor) is False: #this preserves scaling at the cost of vertical resolution
            if max_inst < 4095:
                print("CAN NOT PRESERVE WF OVER HALF OF RESOLUTION IS GONE")
                scale_factor = 8191/max_abs #will not preserve scaling when rounding to ints
                break
            scale_factor = max_inst/max_abs
            max_inst -= 1
    scaled_data = data*scale_factor
    total = 8191*2 + 1
    loss = 100* (abs(np.max(scaled_data)) + abs(np.min(scaled_data)))/total
    print("Estimated Peak-to-Peak Ratio of targetted value is {:.1f}%".format(loss))
    return scaled_data.astype(np.int32)


"""
Error handling: maybe make a seperate python file to take care of error handling 

Also good idea to add option to suppress warnings, aka no print statements instead call warning function that has param that can be suppressed
"""

def exit_with_error(msg):
    """
    Function to raise error message that provides faster feedback
    """
    raise ValueError(msg)



import pandas as pd
import numpy as np
import os

### FILE HANDLING CONVINIENCE FUNCTIONS ###

def metadata_and_data_to_csv(metadata, data, path):
    """
    Convenience function that takes two arbitrary dataframes and writes them to a csv one below the other with a space in between.
    Used nominally for a 1XN metadata df and a data df

    :param metadata: dataframe containing metadata, standard is a 1XN table (many columns with one value each)
    :param data: dataframe containing time data captured from measurment, appended to csv below metadata
    :param path: path to save csv in
    """

    metadata.to_csv(path, index=False, header=True)

    # Add a blank line to the CSV file
    with open(path, 'a') as f:
        f.write('\n')

    # Append data to the same CSV file with its own header
    data.to_csv(path, mode='a', index=False, header=True)

def standard_csv_to_metadata_and_data(path, metadata_header_row=0, data_header_row=2):
    """
    Convenience function that takes the piec standard 1xN metadata with data below and returns each as individual dataframes

    :param path: path to save csv in
    :param metadata_header_row: row where metadata starts (defaut row 0)
    :param data_header_row:  row where data starts (defaut row 2)
    """
    # Read metadata using its header row and assuming it has only one row of data
    metadata = pd.read_csv(path, header=metadata_header_row, nrows=1)

    # Read data starting from its header row and continuing to the end of the file
    data = pd.read_csv(path, header=data_header_row)

    return metadata, data

def create_measurement_filename(directory, measurement_type, notes="", type="csv"):
    """
    Creates a unique filename for a measurement file by checking for identical filenames
    and prepending an index to the filename if an identical one already exists in
    the target directory. Will also create the target directory if it does not already exist.
    
    Parameters:
    - directory (str): Target directory where measurement is to be saved.
    - measurement_type (str): Tag that is unique to the measurement type, e.g., "hyst" or "3pp".
    - notes (str): Extra string to add to the end of the filename, e.g., frequency and voltage info.
    - type (str): File extension, defaults to "csv".
    
    Returns:
    - filename (str): Full filepath of the given measurement in the form "{index}_{measurement_type}_{notes}.{type}"
    """
    # Ensure the directory exists
    os.makedirs(directory, exist_ok=True)
    
    # Construct the base filename
    base_filename = f"{measurement_type}_{notes}.{type}"
    
    # Initialize index
    index = 0
    
    # Loop to find a unique filename
    while True:
        # Construct the full filepath
        filename = os.path.join(directory, f"{index}_{base_filename}")
        
        # Check if the file already exists
        if not os.path.exists(filename):
            break
        
        # Increment the index if the file exists
        index += 1
    
    return filename

### ARBITRARY WAVEFORM CONVINIENCE FUNCTIONS ###

def interpolate_sparse_to_dense(x_sparse, y_sparse, total_points=100):
    """
    Transform sparse arrays of x and y coordinates into a dense array of y coordinates
    linearly interpolated over N=total_points evenly-spaced x values.
    
    Parameters:
    - x_sparse (array-like): Sparse array of x coordinates.
    - y_sparse (array-like): Sparse array of y coordinates.
    - total_points (int): Number of interpolated points between each pair of coordinates.
    
    Returns:
    - y_dense (numpy array): Dense array of linearly interpolated y coordinates.
    """
    y_dense = []

    # Iterate through each pair of adjacent sparse points
    for i in range(len(x_sparse) - 1):
        # Get the start and end points
        x_start, x_end = x_sparse[i], x_sparse[i + 1]
        y_start, y_end = y_sparse[i], y_sparse[i + 1]
        
        # Generate interpolated points between y_start and y_end
        n_to_interpolate = int(((x_sparse[i + 1]-x_sparse[i])/max(x_sparse))*total_points)
        y_interp = np.linspace(y_start, y_end, n_to_interpolate, endpoint=False)
        
        # Append the interpolated points
        y_dense.extend(y_interp)

    #add on duplicate points at the end to ensure array length == total_points (make up for int rounding error)
    while len(y_dense) < total_points:
        y_dense.append(y_dense[-1])

    return np.array(y_dense)




'''
This is for the KEYSIGHT DSOX3024a Oscilloscope and requires the KEYSIGHT I/O Libraries to function.
'''
import numpy as np
import time
from ..scpi_instrument import Scope
#yes

class Dsox3024a(Scope):
    """
    Specific Class for this exact model of scope: Keysight DSOX3024a
    """
    #add class attributes here, like max y range etc
    #correct syntax is tuple for ranges, list for limited amount, and dictionaries for nested things...
    voltage_range = (8e-3, 40)
    voltage_scale = (8e-4, 4)
    time_range = (2e-8, 500)
    time_scale = (2e-9, 50)
    time_base_type = ['MAIN', 'WINDow', 'WIND', 'XY', 'ROLL'] #added WIND so either WIND or WINDOW is allowed
    channel = ['1', '2', '3', '4']

    def __class_specific(self):
        """
        Place to define instrument specific stuff. Ideally never needed if Parent class is robust enough and instead
        we just define class attributes as above
        """
        return None

    def setup(self, channel: str = 1, voltage_range: str = 16, voltage_offset: str = 1, delay: str = '100e-6', time_range: str = '1e-3', autoscale=True):
        """
        Override default params here by ovverriding class Scope funtions
        """
        return super().setup(channel, voltage_range, voltage_offset, delay, time_range, autoscale)






import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.signal import find_peaks
from piec.analysis.utilities import *

def process_raw_3pp(path:str, show_plots=False, save_plots=False, auto_timeshift=True):
    metadata, raw_df = standard_csv_to_metadata_and_data(path)
    processed_df = raw_df

    processed_df['time (s)'] = processed_df['time (s)'].values - processed_df['time (s)'].values[0] # make sure time starts at zero

    # get relevant values from metadata and math
    reset_amp = metadata['reset_amp'].values[0]
    reset_width = metadata['reset_width'].values[0]
    reset_delay = metadata['reset_delay'].values[0]
    p_u_amp = metadata['p_u_amp'].values[0]
    p_u_width = metadata['p_u_width'].values[0]
    p_u_delay = metadata['p_u_delay'].values[0]
    timestep = processed_df['time (s)'].values[-1]/len(processed_df)
    area = metadata['area'].values[0]
    length = metadata['length'].values[0]
    time_offset = metadata['time_offset'].values[0]
    polarity = np.sign(p_u_amp) #palarity of the waveform

    # add on time-dependent processed arrays
    processed_df['current (A)'] = processed_df['voltage (V)']/50/area*100 # 50Ohm conversion, area correction, C/m^2 to uC/cm^2
    processed_df['polarization (uC/cm^2)'] = cumulative_trapezoid(processed_df['current (A)'], processed_df['time (s)'], initial=0)

    N_t0 = np.searchsorted(processed_df['time (s)'].values, time_offset) # manual t0 specification

    if auto_timeshift:
        threshold = np.std(processed_df[processed_df['time (s)'] < reset_width+reset_delay]['voltage (V)'].values*polarity)*0.3 # peak threshold is 30% of the RC discharge of the reset pulse
        distance = min([reset_delay, p_u_delay+p_u_width])/timestep*0.9 # peaks should never be closer than the minimum distance betweeen pulses
        peaks, _ = find_peaks(-polarity*processed_df['voltage (V)'], height=threshold, distance=distance)
        try:
            first_peak = peaks[0]
            v_at_first_peak = -polarity*processed_df['voltage (V)'].values[first_peak]
            rc_rise = 0
            for i in range(first_peak): # want to start pulse not at the peak voltage but at the beginning of the rise to the peak voltage
                if -polarity*processed_df['voltage (V)'].values[first_peak-i] < v_at_first_peak*0.1:
                    rc_rise = i
                    break
            N_t0 = first_peak - rc_rise
            time_offset = N_t0*timestep
        except:
            print('WARNING:INITIAL PEAK NOT FOUND, DEFAULTING TO MANUAL TIME OFFSET CORRECTION')

        processed_df['time (s)'] = processed_df['time (s)'].values - processed_df['time (s)'].values[N_t0] # zero time correction

    # time to choppy chop the pund based on the extracted pulse widths and delays
    n_ph = np.searchsorted(processed_df['time (s)'].values, reset_width+reset_delay)
    n_phr = np.searchsorted(processed_df['time (s)'].values, reset_width+reset_delay+p_u_width)
    n_ps = np.searchsorted(processed_df['time (s)'].values, reset_width+reset_delay+p_u_width+p_u_delay)
    n_psr = np.searchsorted(processed_df['time (s)'].values, reset_width+reset_delay+2*p_u_width+p_u_delay)
    try:
        n_end = np.searchsorted(processed_df['time (s)'].values, reset_width+reset_delay+2*p_u_width+2*p_u_delay)
    except:
        n_end = len(processed_df)

    ph = processed_df['polarization (uC/cm^2)'].values[n_ph:n_phr]
    phr = processed_df['polarization (uC/cm^2)'].values[n_phr:n_ps]
    ps = processed_df['polarization (uC/cm^2)'].values[n_ps:n_psr]
    psr = processed_df['polarization (uC/cm^2)'].values[n_psr:n_end]

    # homogenize lengths for array math
    ph = ph[:min(len(ph), len(ps))]
    ps = ps[:min(len(ph), len(ps))]
    phr = phr[:min(len(phr), len(psr))]
    psr = psr[:min(len(phr), len(psr))]

    dp = np.concatenate([ph, phr]) - np.concatenate([ps, psr]) # time dependent FE polarization is diff between p and u pulse polarizations
    array_dict = {'P^':ph, 'P*':ps, 'P^r':phr, 'P*r':psr, 'dP':dp} # this naming convention mimics the one set by Radiant

    for key in array_dict.keys():
        array_dict[key] -= array_dict[key][0] # zero polarizations
        repeat_values = np.zeros(len(processed_df)-len(array_dict[key]))+array_dict[key][-1]
        array_dict[key] = np.concatenate([array_dict[key], repeat_values]) # add repeat values to the end of arrays so they can be added to the dataframe
        processed_df[key+' (uC/cm^2)'] = array_dict[key] # add analysys arrays to dataframe

    # create applied voltage array from nominal assumptions
    times = [0, reset_width, reset_delay, p_u_width, p_u_delay, p_u_width, p_u_delay,]
    sum_times = [sum(times[:i+1]) for i, t in enumerate(times)]
    # calculate full amplitude of pulse profile and fractional amps of pulses
    amplitude = abs(reset_amp) + abs(p_u_amp)
    frac_reset_amp = reset_amp/amplitude
    frac_p_u_amp = p_u_amp/amplitude

    # specify sparse t and v coordinates which define PUND pulse train
    sparse_t = np.array([sum_times[0], sum_times[1], sum_times[1], sum_times[2], sum_times[2], sum_times[3], sum_times[3],
                            sum_times[4], sum_times[4], sum_times[5], sum_times[5], sum_times[6],])
    sparse_v = np.array([-frac_reset_amp, -frac_reset_amp, 0, 0, frac_p_u_amp, frac_p_u_amp, 0, 0,
                            frac_p_u_amp, frac_p_u_amp, 0, 0,]) * polarity
    
    n_points = int(length/timestep) # n points to use is max

    # densify the array, rise/fall times of pulses will be equal to the awg resolution
    v_applied = interpolate_sparse_to_dense(sparse_t, sparse_v, total_points=n_points)
    initial_delay = np.zeros(int(time_offset//timestep))
    v_applied = np.concatenate([initial_delay, v_applied])

    if len(v_applied)<len(processed_df):
        v_applied = np.concatenate([v_applied, np.zeros(len(processed_df) - len(v_applied))]) # make sure arrays are the same length

    processed_df['applied voltage (V)'] = v_applied[:len(processed_df)]

    processed_df['time (s)'] = processed_df['time (s)'].values - processed_df['time (s)'].values[0] # make sure time starts at zero again

    # optional plotting
    if show_plots or save_plots:
        #dP vs time plot
        fig, ax = plt.subplots(tight_layout=True)
        ax.plot(processed_df['time (s)'], processed_df['dP (uC/cm^2)'], color='k')
        ax.set_xlabel('time (s)')
        ax.set_ylabel('dP (uC/cm^2)')
        if save_plots:
            fig.savefig(path[:-4]+'_dPvst.png')
        if show_plots:
            plt.show()
        plt.close()

        #Current response and applied voltage trace plot
        fig, ax = plt.subplots(tight_layout=True)
        ax.plot(processed_df['time (s)'], processed_df['current (A)'], color='k')
        ax.set_xlabel('time (s)')
        ax.set_ylabel('current (A)')
        ax1 = ax.twinx()
        ax1.plot(processed_df['time (s)'], processed_df['applied voltage (V)'], color='r')
        ax1.set_ylabel('applied voltage (V)')
        if save_plots:
            fig.savefig(path[:-4]+'_trace.png')
        if show_plots:
            plt.show()
        plt.close()

    metadata['time_offset'] = time_offset
    metadata['processed'] = True
    # update csv with new processed data
    metadata_and_data_to_csv(metadata, processed_df, path)